import numpy as np
import itertools
from sklearn.mixture import GaussianMixture
from FreeTrace.module.trajectory_object import TrajectoryObj


def post_processing(trajectory_list, cutoff, verbose=0):
    gap_distrib = []
    filtered_traj_list = []
    post_processed_count = 0
    traj_index = 0
    
    for traj in trajectory_list:
        pos = traj.get_positions()
        if len(pos) >= cutoff:
            frames = traj.get_times()
            xs = pos[:, 0]
            ys = pos[:, 1]

            frame_sorted_args = np.argsort(frames)
            xs = xs[frame_sorted_args]
            ys = ys[frame_sorted_args]
            frames = frames[frame_sorted_args]

            pos = np.vstack([xs, ys]).T
            predicted_false_change_pair = []

            if len(xs) >= 5:
                #disps = np.sqrt((xs[1:] - xs[:-1])**2 + (ys[1:] - ys[:-1])**2)
                if len(xs) <= 10:
                    n_comp = 2
                elif len(xs) < 100:
                    n_comp = 3
                else:
                    n_comp = 4

                gm = GaussianMixture(n_components=n_comp, max_iter=1000, init_params='k-means++', covariance_type='spherical')
                gm = gm.fit(pos)
                labels = gm.predict(pos)
                change_label={}
                pairs = list(itertools.combinations(range(n_comp), r=2))
                for pair in pairs:
                    pair_distance = np.sqrt(np.sum((gm.means_[pair[0]] - gm.means_[pair[1]])**2))
                    if pair_distance < 1.5:
                        change_label[pair[1]] = pair[0]
                for key in sorted(list(change_label.keys()))[::-1]:
                    for idx in range(len(labels)):
                        if labels[idx] == key:
                            labels[idx] = change_label[key]
                means_for_labels = np.array([gm.means_[label] for label in labels])
                
                disps_for_means = np.sqrt(np.sum((pos - means_for_labels)**2, axis=1))

                gaps = []
                for label in sorted(np.unique(labels)):
                    check = np.array([idx for idx, lb in enumerate(labels) if lb==label])
                    gaps.append(np.mean(disps_for_means[check]))
                #gap /= len(disps_for_means)
                gap_distrib.append(np.max(gaps))
    gap_distrib = np.array(gap_distrib)
    if len(gap_distrib[gap_distrib < 1.0])/len(gap_distrib) > 0.8:
        seuil = 1.0
    else:
        seuil = 0.0

    for traj in trajectory_list:
        pos = traj.get_positions()
        if len(pos) >= cutoff:
            frames = traj.get_times()
            xs = pos[:, 0]
            ys = pos[:, 1]

            frame_sorted_args = np.argsort(frames)
            xs = xs[frame_sorted_args]
            ys = ys[frame_sorted_args]
            frames = frames[frame_sorted_args]

            pos = np.vstack([xs, ys]).T
            predicted_false_change_pair = []

            if len(xs) >= 5:
                #disps = np.sqrt((xs[1:] - xs[:-1])**2 + (ys[1:] - ys[:-1])**2)
                if len(xs) <= 10:
                    n_comp = 2
                elif len(xs) < 100:
                    n_comp = 3
                else:
                    n_comp = 4

                gm = GaussianMixture(n_components=n_comp, max_iter=1000, init_params='k-means++', covariance_type='spherical')
                gm = gm.fit(pos)
                labels = gm.predict(pos)
                change_label={}
                pairs = list(itertools.combinations(range(n_comp), r=2))
                for pair in pairs:
                    pair_distance = np.sqrt(np.sum((gm.means_[pair[0]] - gm.means_[pair[1]])**2))
                    if pair_distance < 1.5:
                        change_label[pair[1]] = pair[0]
                for key in sorted(list(change_label.keys()))[::-1]:
                    for idx in range(len(labels)):
                        if labels[idx] == key:
                            labels[idx] = change_label[key]
                means_for_labels = np.array([gm.means_[label] for label in labels])
                
                disps_for_means = np.sqrt(np.sum((pos - means_for_labels)**2, axis=1))

                gap = 0
                for label in sorted(np.unique(labels)):
                    check = np.array([idx for idx, lb in enumerate(labels) if lb==label])
                    gap += np.sum(disps_for_means[check])
                gap /= len(disps_for_means)

                unique_labels = sorted(np.unique(labels))
                pairs = list(itertools.combinations(unique_labels, r=2))
                change_label_distances = {(pair[0], pair[1]):[] for pair in pairs}
                prev_label = labels[0]
                for idx in range(1, len(pos)):
                    cur_label = labels[idx]
                    if cur_label != prev_label:
                        a, b = min(prev_label, cur_label), max(prev_label, cur_label)
                        change_label_distances[(a, b)].append(np.sqrt(np.sum((pos[idx] - pos[idx-1])**2)))
                    prev_label = cur_label
    
                for pair in pairs:
                    changing_distances = change_label_distances[(pair[0], pair[1])]
                    if gap < seuil:
                        predicted_false_change_pair.append(pair)


            if len(predicted_false_change_pair) > 0:
                post_processed_count += 1
                prev_label = labels[0]
                new_traj = TrajectoryObj(index=traj_index)
                new_traj.add_trajectory_position(frames[0], xs[0], ys[0], 0)
                for lb_idx in range(1, len(labels)):
                    cur_label = labels[lb_idx]
                    if (min(prev_label, cur_label), max(prev_label, cur_label)) in predicted_false_change_pair:
                        if len(new_traj.get_positions()) >= cutoff:
                            filtered_traj_list.append(new_traj)
                            traj_index += 1
                        new_traj = TrajectoryObj(index=traj_index)
                    new_traj.add_trajectory_position(frames[lb_idx], xs[lb_idx], ys[lb_idx], 0)
                    prev_label = cur_label
                if len(new_traj.get_positions()) >= cutoff:
                    filtered_traj_list.append(new_traj)
                    traj_index += 1
            else:
                new_traj = TrajectoryObj(index=traj_index)
                for idx in range(len(xs)):
                    new_traj.add_trajectory_position(frames[idx], xs[idx], ys[idx], 0)
                if len(new_traj.get_positions()) >= cutoff:
                    filtered_traj_list.append(new_traj)
                    traj_index += 1


    if verbose != 0:
        print(f'Post processed nb of trajectories: {post_processed_count} with ratio: {len(gap_distrib[gap_distrib < 1.0])/len(gap_distrib)}, seuil: {seuil}')
    return filtered_traj_list


def post_processing2(trajectory_list, cutoff, verbose=0):
    gap_distrib = []
    filtered_traj_list = []
    post_processed_count = 0
    traj_index = 0
    
    for traj in trajectory_list:
        pos = traj.get_positions()
        if len(pos) >= cutoff:
            frames = traj.get_times()
            xs = pos[:, 0]
            ys = pos[:, 1]

            frame_sorted_args = np.argsort(frames)
            xs = xs[frame_sorted_args]
            ys = ys[frame_sorted_args]
            frames = frames[frame_sorted_args]

            pos = np.vstack([xs, ys]).T
            predicted_false_change_pair = []

            if len(xs) >= 5:
                #disps = np.sqrt((xs[1:] - xs[:-1])**2 + (ys[1:] - ys[:-1])**2)
                if len(xs) <= 10:
                    n_comp = 2
                elif len(xs) < 100:
                    n_comp = 3
                else:
                    n_comp = 4

                gm = GaussianMixture(n_components=n_comp, max_iter=1000, init_params='k-means++', covariance_type='spherical')
                gm = gm.fit(pos)
                labels = gm.predict(pos)
                change_label={}
                pairs = list(itertools.combinations(range(n_comp), r=2))
                for pair in pairs:
                    pair_distance = np.sqrt(np.sum((gm.means_[pair[0]] - gm.means_[pair[1]])**2))
                    if pair_distance < 1.5:
                        change_label[pair[1]] = pair[0]
                for key in sorted(list(change_label.keys()))[::-1]:
                    for idx in range(len(labels)):
                        if labels[idx] == key:
                            labels[idx] = change_label[key]
                means_for_labels = np.array([gm.means_[label] for label in labels])
                
                disps_for_means = np.sqrt(np.sum((pos - means_for_labels)**2, axis=1))

                gaps = []
                for label in sorted(np.unique(labels)):
                    check = np.array([idx for idx, lb in enumerate(labels) if lb==label])
                    gaps.append(np.mean(disps_for_means[check]))
                #gap /= len(disps_for_means)
                gap_distrib.append(np.max(gaps))
    gap_distrib = np.array(gap_distrib)
    if len(gap_distrib[gap_distrib < 1.0])/len(gap_distrib) > 0.8:
        seuil = 1.0
    else:
        seuil = 0.0

    for traj in trajectory_list:
        pos = traj.get_positions()
        if len(pos) >= cutoff:
            frames = traj.get_times()
            xs = pos[:, 0]
            ys = pos[:, 1]

            frame_sorted_args = np.argsort(frames)
            xs = xs[frame_sorted_args]
            ys = ys[frame_sorted_args]
            frames = frames[frame_sorted_args]

            pos = np.vstack([xs, ys]).T
            predicted_false_change_pair = []

            if len(xs) >= 5:
                disps = np.sqrt((xs[1:] - xs[:-1])**2 + (ys[1:] - ys[:-1])**2)
                if len(xs) <= 10:
                    n_comp = 5
                elif len(xs) < 100:
                    n_comp = 8
                else:
                    n_comp = 10

                gm = GaussianMixture(n_components=n_comp, n_init=3, max_iter=1000, init_params='k-means++', covariance_type='tied')
                gm = gm.fit(pos)
                labels = gm.predict(pos)
                change_label={}
                pairs = list(itertools.combinations(range(n_comp), r=2))
                for pair in pairs:
                    pair_distance = np.sqrt(np.sum((gm.means_[pair[0]] - gm.means_[pair[1]])**2))
                    if pair_distance < 0.5:
                        change_label[pair[1]] = pair[0]
                for key in sorted(list(change_label.keys()))[::-1]:
                    for idx in range(len(labels)):
                        if labels[idx] == key:
                            labels[idx] = change_label[key]
                means_for_labels = np.array([gm.means_[label] for label in labels])
                
                disps_for_means = np.sqrt(np.sum((pos - means_for_labels)**2, axis=1))

                gap = 0
                gaps = []
                for label in sorted(np.unique(labels)):
                    check = np.array([idx for idx, lb in enumerate(labels) if lb==label])
                    gap += np.sum(disps_for_means[check])
                    gaps.append(np.mean(disps_for_means[check]))
                gap /= len(disps_for_means)

                unique_labels = sorted(np.unique(labels))
                pairs = list(itertools.combinations(unique_labels, r=2))
                change_label_distances = {(pair[0], pair[1]):[] for pair in pairs}
                prev_label = labels[0]

                fig, axs = plt.subplots(1, 4, figsize=(16, 4))
                for idx in range(1, len(pos)):
                    cur_label = labels[idx]
                    if cur_label != prev_label:
                        a, b = min(prev_label, cur_label), max(prev_label, cur_label)
                        change_label_distances[(a, b)].append(np.sqrt(np.sum((pos[idx] - pos[idx-1])**2)))
                    prev_label = cur_label
                    axs[0].plot([pos[idx-1][0], pos[idx][0]], [pos[idx-1][1], pos[idx][1]])
    
                for pair in pairs:
                    changing_distances = change_label_distances[(pair[0], pair[1])]
                    if gap < seuil:
                        predicted_false_change_pair.append(pair)
                
                print('GAP: ', gap, gaps)
                print(labels)
                axs[0].scatter(pos[:,0], pos[:,1])
                axs[0].scatter(gm.means_[:,0], gm.means_[:,1], c='red')
                window_width = max(np.max(pos[:, 0]) - np.min(pos[:, 0]), np.max(pos[:, 0]) - np.min(pos[:, 0])) + 4
                xcenter = np.mean(pos[:, 0])
                ycenter = np.mean(pos[:, 1])
                axs[0].set_xlim([xcenter - window_width//2, xcenter + window_width//2])
                axs[0].set_ylim([ycenter - window_width//2, ycenter + window_width//2])
                axs[1].hist(disps, bins=np.linspace(0, 10, 80))

                clf = LocalOutlierFactor(n_neighbors=5)
                clf.fit_predict(pos)
                print(clf.negative_outlier_factor_)
                print(disps)
                axs[2].hist(clf.negative_outlier_factor_, bins=np.linspace(min(-10, np.min(clf.negative_outlier_factor_)), 0, 20))

                """
                G = nx.Graph()
                for i in range(0, len(pos)):
                    G.add_node(i, position=pos[i])
                for i in range(0, len(pos)):
                    for j in range(i+1, len(pos)):
                        #G.add_edge(i, j, distance=1 / np.sqrt(np.sum((pos[i] - pos[j])**2))**2)
                        G.add_edge(i, j, distance=1)
                    #G.add_edge(idx-1, idx)
                communs = nx.community.greedy_modularity_communities(G, weight='distance', resolution=1, best_n=3, cutoff=3)
                for commun, color in zip(communs, ['red', 'orange', 'yellow', 'green', 'blue', 'cyan', 'purple', 'gray', 'black']):
                    for node in commun:
                        x, y = G.nodes[node]['position']
                        axs[3].scatter(x, y, c=color)
                """

                plt.show()

            if len(predicted_false_change_pair) > 0:
                post_processed_count += 1
                prev_label = labels[0]
                new_traj = TrajectoryObj(index=traj_index)
                new_traj.add_trajectory_position(frames[0], xs[0], ys[0], 0)
                for lb_idx in range(1, len(labels)):
                    cur_label = labels[lb_idx]
                    if (min(prev_label, cur_label), max(prev_label, cur_label)) in predicted_false_change_pair:
                        if len(new_traj.get_positions()) >= cutoff:
                            filtered_traj_list.append(new_traj)
                            traj_index += 1
                        new_traj = TrajectoryObj(index=traj_index)
                    new_traj.add_trajectory_position(frames[lb_idx], xs[lb_idx], ys[lb_idx], 0)
                    prev_label = cur_label
                if len(new_traj.get_positions()) >= cutoff:
                    filtered_traj_list.append(new_traj)
                    traj_index += 1
            else:
                new_traj = TrajectoryObj(index=traj_index)
                for idx in range(len(xs)):
                    new_traj.add_trajectory_position(frames[idx], xs[idx], ys[idx], 0)
                if len(new_traj.get_positions()) >= cutoff:
                    filtered_traj_list.append(new_traj)
                    traj_index += 1


    if verbose != 0:
        print(f'Post processed nb of trajectories: {post_processed_count} with ratio: {len(gap_distrib[gap_distrib < 1.0])/len(gap_distrib)}, seuil: {seuil}')
    return filtered_traj_list


"""
import networkx as nx
from sklearn.neighbors import LocalOutlierFactor
from FreeTrace.module.data_load import read_csv
import matplotlib.pyplot as plt
my_traj_list = []
#df = read_csv('outputs/20250318-RPE-1-FUS-1C9-Wang4-100uM-JF200nM-Hoetchst03-4H-cell24-IR10min_immobile_dense_traces_postoff_0.5.csv')
#df = read_csv('outputs/20231129-cell6-ir_whole_traces_postoff_0.5.csv')
#df = read_csv('outputs/20231129-cell6-whole-noIR_traces_postoff_0.5.csv')
df = read_csv('outputs/alpha_test0_traces_postoff_0.5.csv')
#df = read_csv('outputs/alpha_test1_traces_postoff_0.5.csv')
#df = read_csv('outputs/sample0_traces_postoff_0.5.csv')
#df = read_csv('outputs/20250219-RPE-1-WT-1C9-JF200nM-Hoetchst03-cell3-IR10min_traces_postoff_0.5.csv')
#df = read_csv('outputs/20250219-RPE-1-WT-1C9-JF200nM-Hoetchst03-cell2_traces_postoff_0.5.csv')
traj_indices = df['traj_idx'].unique()
for ix, traj_idx in enumerate(traj_indices):
    single_traj = df[df['traj_idx'] == traj_idx]
    xs = single_traj['x'].to_numpy()
    ys = single_traj['y'].to_numpy()
    frames = single_traj['frame'].to_numpy()
    traj = TrajectoryObj(index=ix)
    for x, y, frame in zip(xs, ys, frames):
        traj.add_trajectory_position(frame, x, y, 0)
    my_traj_list.append(traj)
post_processing2(my_traj_list, 3, 1)
exit()
"""


"""
def post_processing(trajectory_list, cutoff):
    length_check = 3
    std_expectation_cut = 0.075

    filtered_pos = []
    filtered_frames = []
    for traj in trajectory_list:
        start_n_comp = 3
        delete_idx = []

        pos = traj.get_positions()
        frames = traj.get_times()
        xs = pos[:, 0]
        ys = pos[:, 1]
        zs = pos[:, 2]

        frame_sorted_args = np.argsort(frames)
        xs = xs[frame_sorted_args]
        ys = ys[frame_sorted_args]
        zs = zs[frame_sorted_args]
        frames = frames[frame_sorted_args]
        
        if len(xs) >= 5:
            disps = np.sqrt((xs[1:] - xs[:-1])**2 + (ys[1:] - ys[:-1])**2).reshape((len(xs) - 1), 1)
            if np.std(disps) < 1e-6:
                continue
            while True:
                if start_n_comp == 1:
                    break
                flag = 0
                gm = GaussianMixture(start_n_comp, max_iter=100, init_params='k-means++').fit(disps / np.max(disps))
                means_ = gm.means_.flatten()
                stds_ = gm.covariances_.flatten()
                pairs = itertools.combinations(range(start_n_comp), r=2)
                
                for pair in pairs:
                    if abs(means_[pair[0]] - means_[pair[1]]) < 0.5 \
                        or (means_[pair[1]] - 4*stds_[pair[1]] < means_[pair[0]] + 4*stds_[pair[0]] < means_[pair[1]] + 4*stds_[pair[1]]) \
                        and (means_[pair[0]] - 4*stds_[pair[0]] < means_[pair[1]] + 4*stds_[pair[1]] < means_[pair[0]] + 4*stds_[pair[0]]):
                        start_n_comp -= 1
                        flag = 1
                        break
                if flag == 0:
                    break
            
            if start_n_comp >= 2:         
                labels = gm.predict(disps / np.max(disps))
                expects = [means_[label] for label in labels]
                std_for_class = ((disps / np.max(disps)).flatten() - expects)**2
                std_expect = np.sqrt(np.mean(std_for_class))
                if std_expect < std_expectation_cut:
                    label_count = [0 for _ in np.unique(labels)]
                    i = 1
                    before_label = labels[0]
                    cuts = []

                    while True:
                        cur_label = labels[i]
                        if before_label != cur_label:
                            cuts.append(i)
                        i+=1
                        if i == len(labels):
                            break

                    before_label = labels[0]
                    chunk_idx = [0, len(labels)]
                    for lb_idx, label in enumerate(labels):
                        if label != before_label:
                            chunk_idx.append(lb_idx)
                        before_label = label
                    chunk_idx = sorted(chunk_idx)
                    for label in labels:
                        label_count[label] += 1
                    max_label = np.argmax(label_count)

                    for idx in range(len(chunk_idx) - 1):
                        if (chunk_idx[idx+1] - chunk_idx[idx]) <= length_check and labels[chunk_idx[idx]] != max_label:
                            delete_idx.extend(list(range(chunk_idx[idx], chunk_idx[idx+1])))
                print(delete_idx, disps, std_expect)
        if len(delete_idx) > 0:
            new_xs = []
            new_ys = []
            new_zs = []
            new_frames = []

            delete_idx = np.array(delete_idx)
            for i in range(len(xs)):
                if i in delete_idx:
                    new_xs.append(xs[i])
                    new_ys.append(ys[i])
                    new_zs.append(zs[i])
                    new_frames.append(frames[i])
                    if len(new_xs) >= 2:
                        filtered_pos.append([new_xs, new_ys, new_zs])
                        filtered_frames.append(new_frames)
                    new_xs = []
                    new_ys = []
                    new_zs = []
                    new_frames = []
                else:
                    new_xs.append(xs[i])
                    new_ys.append(ys[i])
                    new_zs.append(zs[i])
                    new_frames.append(frames[i])
        else:
            filtered_pos.append([xs, ys, zs])
            filtered_frames.append(frames)
            
    filtered_trajectory_list = []
    traj_idx = 0
    for path, frames in zip(filtered_pos, filtered_frames):
        if len(path) >= cutoff:
            traj = TrajectoryObj(index=traj_idx)
            for node_idx in range(len(frames)):
                x = path[0][node_idx]
                y = path[1][node_idx]
                z = path[2][node_idx]
                frame = frames[node_idx]
                traj.add_trajectory_position(frame, x, y, z)
            filtered_trajectory_list.append(traj)
            traj_idx += 1
    return filtered_trajectory_list
"""



import os
import csv
import numpy as np
import pandas as pd


def save_report(data: list, path='', all=False) -> list:
    """
    @params : data(list), path(String), all(boolean)
    @return : list of saved report names(list)
    Save reports in a given path.
    """
    histones = {}
    report_names = []

    for chunked_data in data:
        histones |= chunked_data

    if not all:
        histone_names = list(histones.keys())
        filenames = set()
        for histone in histone_names:
            filenames.add(histone.split('\\')[-1].split('@')[0])

        for filename in filenames:
            h = {}
            for histone in histone_names:
                if filename in histone:
                    h[histone] = histones[histone]
            write_file_name = f'{path}/{filename}.csv'
            with open(write_file_name, 'w', newline='') as f:
                report_names.append(write_file_name)
                fieldnames = ['filename', 'h2b_id', 'predicted_class_id', 'predicted_class_name', 'probability',
                              'maximum_radius', 'first_x_position', 'first_y_position']
                writer = csv.DictWriter(f, fieldnames=fieldnames)
                writer.writeheader()

                for i, key in enumerate(h):
                    trajectory = histones[key].get_trajectory()
                    first_x_pos = trajectory[0][0]
                    first_y_pos = trajectory[0][1]
                    file_name = histones[key].get_file_name()
                    h2b_id = histones[key].get_id()
                    pred_class_id = histones[key].get_predicted_label()
                    max_r = histones[key].get_max_radius()
                    proba = histones[key].get_predicted_proba()

                    pred_class_name = 'unidentified'
                    if pred_class_id == 0:
                        pred_class_name = 'Immobile'
                    if pred_class_id == 1:
                        pred_class_name = 'Hybrid'
                    if pred_class_id == 2:
                        pred_class_name = 'Mobile'

                    writer.writerow({'filename':file_name, 'h2b_id':h2b_id, 'predicted_class_id':pred_class_id,
                                     'predicted_class_name':pred_class_name, 'probability':proba, 'maximum_radius':max_r,
                                     'first_x_position':first_x_pos, 'first_y_position':first_y_pos})
    else:
        write_file_name = f'{path}/prediction_all.csv'
        with open(write_file_name, 'w', newline='') as f:
            report_names.append(write_file_name)
            fieldnames = ['filename', 'h2b_id', 'predicted_class_id', 'predicted_class_name', 'probability',
                          'maximum_radius', 'first_x_position', 'first_y_position']
            writer = csv.DictWriter(f, fieldnames=fieldnames)
            writer.writeheader()

            for i, key in enumerate(histones):
                trajectory = histones[key].get_trajectory()
                first_x_pos = trajectory[0][0]
                first_y_pos = trajectory[0][1]
                file_name = histones[key].get_file_name()
                h2b_id = histones[key].get_id()
                pred_class_id = histones[key].get_predicted_label()
                max_r = histones[key].get_max_radius()
                proba = histones[key].get_predicted_proba()

                pred_class_name = 'unidentified'
                if pred_class_id == 0:
                    pred_class_name = 'Immobile'
                if pred_class_id == 1:
                    pred_class_name = 'Hybrid'
                if pred_class_id == 2:
                    pred_class_name = 'Mobile'

                writer.writerow({'filename': file_name, 'h2b_id': h2b_id, 'predicted_class_id': pred_class_id,
                                 'predicted_class_name': pred_class_name, 'probability': proba, 'maximum_radius': max_r,
                                 'first_x_position': first_x_pos, 'first_y_position': first_y_pos})
    return report_names


def write_model_info(training_model, path: str, history: list, nb_histones: int, date: str) -> str:
    """
    @params : model(tensorflow model object), path(String), history(list), nb_histones(Integer), data(String)
    @return : saved model name(String)
    Save the logs and the history of the model and return the model name.
    """
    new_model_num = 0
    try:
        if os.path.isdir(path):
            contents = os.listdir(path)
            for content in contents:
                if 'model' in content:
                    model_num = int(content.split('_')[0].split('model')[-1])
                    new_model_num = max(new_model_num, model_num)
            modelname = f'model{new_model_num + 1}'
        training_model.save(f'{path}/{modelname}')
    except Exception as e:
        print('Model directory creation err')
        print(e)

    with open(f'{path}/{modelname}/log.txt', 'w') as info_file:
        info_file.write(f'{date}, number of trained h2bs:{str(nb_histones)}\n')
        info_file.write(f'train history, test history, train_acc, test_acc\n')
        for line_num in range(len(history[0])):
            info_file.write(f'{str(history[0][line_num])}\t{str(history[1][line_num])}\t'
                            f'{str(history[2][line_num])}\t{str(history[3][line_num])}\n')
    return modelname


def write_trxyt(file: str, trajectory_list: list, pixel_microns=1.0, frame_rate=1.0):
    try:
        with open(file, 'w', encoding="utf-8") as f:
            input_str = ''
            for index, trajectory_obj in enumerate(trajectory_list):
                for (xpos, ypos, zpos), time in zip(trajectory_obj.get_positions(), trajectory_obj.get_times()):
                    input_str += f'{index}\t{xpos * pixel_microns:.5f}\t{ypos * pixel_microns:.5f}\t{time * frame_rate:.3f}\n'
            f.write(input_str)
    except Exception as e:
        print(f"Unexpected error, check the file: {file}")
        print(e)


def write_trxyt(file: str, trajectory_list: list, pixel_microns=1.0, frame_rate=1.0):
    try:
        with open(file, 'w', encoding="utf-8") as f:
            input_str = ''
            for index, trajectory_obj in enumerate(trajectory_list):
                for (xpos, ypos, zpos), time in zip(trajectory_obj.get_positions(), trajectory_obj.get_times()):
                    input_str += f'{index}\t{xpos * pixel_microns:.5f}\t{ypos * pixel_microns:.5f}\t{time * frame_rate:.3f}\n'
            f.write(input_str)
    except Exception as e:
        print(f"Unexpected error, check the file: {file}")
        print(e)


def write_localization(output_dir, coords, all_pdfs, infos):
    lines = f'frame,x,y,z,xvar,yvar,rho,norm_cst,intensity,window_size\n'
    for frame, (coord, pdfs, info) in enumerate(zip(coords, all_pdfs, infos)):
        for pos, (x_var, y_var, rho, amp), pdf in zip(coord, info, pdfs):
            window_size = int(np.sqrt(len(pdf)))
            peak_val = pdf[int((len(pdf) - 1) / 2)]
            lines += f'{frame + 1}'
            if len(pos) == 3:
                lines += f',{pos[1]},{pos[0]},{pos[2]}'
            elif len(pos) == 2:
                lines += f',{pos[1]},{pos[0]},0.0'
            elif len(pos) == 1:
                lines += f',{pos[0]},0.0,0.0'
            else:
                print(f'Localization writing Err')
                raise Exception
            lines += f',{x_var},{y_var},{rho},{amp},{peak_val},{window_size}'
            lines += f'\n'

    with open(f'{output_dir}_loc.csv', 'w') as f:
        f.write(lines)


def rewrite_localization(output_dir, coords, infos):
    lines = f"frame,x,y,z,xvar,yvar,rho,norm_cst,intensity,window_size\n"
    for frame in sorted(coords.keys()):
        coord = coords[frame]
        info = infos[frame]
        for (x, y, z), (x_var, y_var, rho, norm_cst, intensity, winsize) in zip(coord, info):
            lines += f"{frame}"
            lines += f",{x},{y},{z}"
            lines += f",{x_var},{y_var},{rho},{norm_cst},{intensity},{winsize}"
            lines += f"\n"
    with open(f"{output_dir}", 'w') as f:
        f.write(lines)


def write_trajectory(file: str, trajectory_list: list):
    try:
        with open(file, 'w', encoding="utf-8") as f:
            input_str = 'traj_idx,frame,x,y,z\n'
            for trajectory_obj in trajectory_list:
                for (xpos, ypos, zpos), time in zip(trajectory_obj.get_positions(), trajectory_obj.get_times()):
                    input_str += f'{trajectory_obj.get_index()},{time},{xpos},{ypos},{zpos}\n'
            f.write(input_str)
    except Exception as e:
        print(f"Unexpected error, check the file: {file}")
        print(e)


def write_H_and_K(filepath: str, H_K_dict: dict):
    H_K_dict = pd.DataFrame(H_K_dict)
    H_K_dict.to_csv(filepath, index=False) 
    


import numpy as np
from FreeTrace.module.trajectory_object import TrajectoryObj as trajobj


def write_xml(output_file: str, trajectory_list: list, snr='4',
              density='low', scenario='VESICLE', cutoff=0):
    input_str = '<?xml version="1.0" encoding="UTF-8" standalone="no"?>\n'
    input_str += '<root>\n'
    input_str += (f'<TrackContestISBI2012 SNR="{snr}" density="{density}" '
                  f'generationDateTime="Mon Nov 20 09:00:00 CET 2023" '
                  f'info="http://bioimageanalysis.org/track/" scenario="{scenario}">\n')
    with open(output_file, 'w', encoding="utf-8") as f:
        for trajectory_obj in trajectory_list:
            if trajectory_obj.get_trajectory_length() >= cutoff:
                input_str += '<particle>\n'
                for (xpos, ypos, zpos), time in zip(trajectory_obj.get_positions(), trajectory_obj.get_times()):
                    input_str += f'<detection t="{time - 1}" x="{xpos}" y="{ypos}" z="{zpos}"/>\n'
                input_str += '</particle>\n'
        input_str += '</TrackContestISBI2012>\n'
        input_str += '</root>\n'
        f.write(input_str)


def xml_to_object(input_file):
    obj_index = 0
    trajectory_list = []
    with open(input_file, 'r', encoding="utf-8") as f:
        lines = f.readlines()[3:]
        for line in lines:
            l = line.split('\n')[0]
            if l == '<particle>':
                trajectory_list.append(trajobj(index=obj_index, max_pause=5))
            elif l == '</particle>':
                obj_index += 1
            elif 'detection' in l:
                c = l.split('\"')
                t, x, y, z = int(c[1]) + 1, float(c[3]), float(c[5]), float(c[7])
                trajectory_list[obj_index].add_trajectory_position(t, x, y, z)
    return trajectory_list


def read_xml(input_file):
    localizations = {}
    with open(input_file, 'r', encoding="utf-8") as f:
        lines = f.readlines()[3:]
        for line in lines:
            l = line.split('\n')[0]
            if 'detection' in l:
                c = l.split('\"')
                t, x, y, z = int(c[1]) + 1, float(c[3]), float(c[5]), float(c[7])
                if t in localizations:
                    localizations[t].append([x, y, z])
                else:
                    localizations[t] = [[x, y, z]]
    return localizations


def trxyt_to_xml(input_file: str, output_file: str, cutoff=0):
    filetypes = ['trxyt', 'trx']
    trajectories = {}
    # Check filetype.
    assert input_file.strip().split('.')[-1].lower() in filetypes
    # Read file and store the trajectory and time information in H2B object
    try:
        with open(input_file, 'r', encoding="utf-8") as f:
            input = f.read()
        lines = input.strip().split('\n')
        for line in lines:
            temp = line.split('\t')
            index = int(temp[0].strip())
            x_pos = float(temp[1].strip())
            y_pos = float(temp[2].strip())
            time_step = int(np.round(((float(temp[3].strip()) * 100) - 1)))
            if index in trajectories:
                trajectories[index].append([x_pos, y_pos, time_step])
            else:
                trajectories[index] = [[x_pos, y_pos, time_step]]
        f.close()

        with open(output_file, 'w', encoding="utf-8") as fxml:
            input = '<?xml version="1.0" encoding="UTF-8" standalone="no"?>\n'
            input += '<root>\n'
            input += ('<TrackContestISBI2012 SNR="7" density="mid" '
                      'generationDateTime="Mon Mar 12 17:20:58 CET 2012" info="http://bioimageanalysis.org/track/" scenario="MICROTUBULE">\n')
            for index in list(trajectories.keys()):
                if len(trajectories[index]) >= cutoff:
                    input += '<particle>\n'
                    for xpos, ypos, t in trajectories[index]:
                        input += f'<detection t="{t}" x="{xpos}" y="{ypos}" z="0"/>\n'
                    input += '</particle>\n'
            input += '</TrackContestISBI2012>\n'
            input += '</root>\n'
            fxml.write(input)
    except Exception as e:
        print(f"Unexpected error, check the file: {input_file}")
        print(e)


def andi_gt_to_xml(input_file: str, output_file: str, cutoff=0):
    filetypes = ['csv']
    trajectories = {}
    # Check filetype.
    assert input_file.strip().split('.')[-1].lower() in filetypes
    # Read file and store the trajectory and time information in H2B object
    try:
        with open(input_file, 'r', encoding="utf-8") as f:
            input = f.read()
        lines = input.strip().split('\n')
        for line in lines[1:]:
            temp = line.split(',')
            index = int(float(temp[0].strip()))
            x_pos = float(temp[2].strip())
            y_pos = float(temp[3].strip())
            time_step = int(float(temp[1].strip()))
            if index in trajectories:
                trajectories[index].append([x_pos, y_pos, time_step])
            else:
                trajectories[index] = [[x_pos, y_pos, time_step]]
        f.close()

        with open(output_file, 'w', encoding="utf-8") as fxml:
            input = '<?xml version="1.0" encoding="UTF-8" standalone="no"?>\n'
            input += '<root>\n'
            input += ('<TrackContestISBI2012 SNR="7" density="mid" '
                      'generationDateTime="Mon Mar 12 17:20:58 CET 2012" info="http://bioimageanalysis.org/track/" scenario="VESICLE">\n')
            for index in list(trajectories.keys()):
                if len(trajectories[index]) >= cutoff:
                    input += '<particle>\n'
                    for xpos, ypos, t in trajectories[index]:
                        input += f'<detection t="{t}" x="{xpos}" y="{ypos}" z="0"/>\n'
                    input += '</particle>\n'
            input += '</TrackContestISBI2012>\n'
            input += '</root>\n'
            fxml.write(input)

    except Exception as e:
        print(f"Unexpected error, check the file: {input_file}")
        print(e)


def mosaic_to_xml(input_file: str, output_file: str, cutoff=0):
    filetypes = ['csv']
    tmp = {}
    # Check filetype.
    assert input_file.strip().split('.')[-1].lower() in filetypes
    # Read file and store the trajectory and time information in H2B object
    try:
        with open(input_file, 'r', encoding="utf-8") as f:
            input = f.read()
        lines = input.strip().split('\n')[1:]
        for line in lines:
            temp = line.split(',')
            x_pos = float(temp[3].strip())
            y_pos = float(temp[4].strip())
            time_step = int(temp[2].strip()) + 1
            obj_index = int(temp[1].strip())
            if obj_index in tmp:
                tmp[obj_index].add_trajectory_position(time_step, x_pos, y_pos, 0)
            else:
                tmp[obj_index] = trajobj(index=obj_index, localizations=None, max_pause=0)
                tmp[obj_index].add_trajectory_position(time_step, x_pos, y_pos, 0)

        objs_list = []
        for obj in tmp:
            objs_list.append(tmp[obj])
        write_xml(output_file=output_file, trajectory_list=objs_list, snr='7', cutoff=cutoff)
    except Exception as e:
        print(f"Unexpected error, check the file: {input_file}")
        print(e)



#cython: infer_types=True
#cython: cdivision=True
from libc.stdlib cimport malloc, free
import numpy as np
cimport cython


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True) 
@cython.profile(False)
cdef float[:,::1] image_overlap(float[:,::1] img1, float[:,::1] img2, int div):
    cdef int row_max, col_max
    cdef Py_ssize_t r, c

    row_max = img1.shape[0]
    col_max = img1.shape[1]

    for r in range(row_max):
        for c in range(col_max):
            img1[r][c] = (img1[r][c] + img2[r][c]) / div
    return img1


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True) 
@cython.profile(False)
cdef float contig_image_mean(float[:,::1] img):
    cdef int count, row_max, col_max
    cdef float sum
    cdef Py_ssize_t r, c

    sum = 0.0
    count = 0
    row_max = img.shape[0]
    col_max = img.shape[1]

    for r in range(row_max):
        for c in range(col_max):
            sum += img[r][c]
            count += 1
    if count != 0:
        return sum / count
    else:
        return 0.0


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True) 
@cython.profile(False)
cdef float image_mean(float[:,:] img):
    cdef int count, row_max, col_max
    cdef float sum
    cdef Py_ssize_t r, c

    sum = 0.0
    count = 0
    row_max = img.shape[0]
    col_max = img.shape[1]

    for r in range(row_max):
        for c in range(col_max):
            sum += img[r][c]
            count += 1
    if count != 0:
        return sum / count
    else:
        return 0.0


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True) 
@cython.profile(False)
cdef float image_std(float[:,:] img):
    cdef int count, row_max, col_max
    cdef float var, mean
    cdef Py_ssize_t r, c

    row_max = img.shape[0]
    col_max = img.shape[1]
    var = 0.0
    count = 0
    mean = image_mean(img)

    for r in range(row_max):
        for c in range(col_max):
            var += (img[r][c] - mean) * (img[r][c] - mean)
            count += 1
    if count != 0:
        return (var / count) ** 0.5
    else:
        return 0.0


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True) 
@cython.profile(False)
cdef float contig_image_std(float[:,::1] img):
    cdef int count, row_max, col_max
    cdef float var, mean
    cdef Py_ssize_t r, c

    row_max = img.shape[0]
    col_max = img.shape[1]
    var = 0.0
    count = 0
    mean = image_mean(img)

    for r in range(row_max):
        for c in range(col_max):
            var += (img[r][c] - mean) * (img[r][c] - mean)
            count += 1
    if count != 0:
        return (var / count) ** 0.5
    else:
        return 0.0


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True) 
@cython.profile(False)
cpdef float[:,::1] boundary_smoothing(float[:,::1] img, int[::1] row_indice, int[::1] col_indice):
    cdef int border_max
    cdef int repeat_n
    cdef int erase_space
    cdef Py_ssize_t border, r, c, i
    cdef int height, width, row_min, row_max, col_min, col_max, index, row, col, row_slice_0, row_slice_1, col_slice_0, col_slice_1

    border_max = 50
    erase_space = 2
    repeat_n = 2
    height = img.shape[0]
    width = img.shape[1]
    index = 0

    cdef float[:, ::1] img_view = img
    cdef int *center_xy = <int *>malloc(border_max * border_max * sizeof(int))

    for border in range(border_max):
        row_min = max(0, row_indice[0] + border)
        row_max = min(height - 1, row_indice[1] - border)
        col_min = max(0, col_indice[0] + border)
        col_max = min(width - 1, col_indice[1] - border)
        for col in range(col_min, col_max+1):
            center_xy[index] = row_min
            index += 1
            center_xy[index] = col
            index += 1
        for row in range(row_min, row_max+1):
            center_xy[index] = row
            index += 1
            center_xy[index] = col_max
            index += 1
        for col in range(col_max, col_min-1, -1):
            center_xy[index] = row_max
            index += 1
            center_xy[index] = col
            index += 1
        for row in range(row_max, row_min-1, -1):
            center_xy[index] = row
            index += 1
            center_xy[index] = col_min
            index += 1

    for _ in range(repeat_n):
        for i in range(index):
            if i % 2 == 0:
                r = center_xy[i]
            else:
                c = center_xy[i]
                row_slice_0 = max(0, r-erase_space)
                row_slice_1 = min(height, r+erase_space+1)
                col_slice_0 = max(0, c-erase_space)
                col_slice_1 = min(width, c+erase_space+1)
                img_view[r][c] = contig_image_mean(img_view[row_slice_0:row_slice_1, col_slice_0:col_slice_1])
    free(center_xy)
    return img


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True) 
@cython.profile(False)
cpdef add_block_noise(imgs, extend):
    gap = extend//2
    row_indice = range(0, len(imgs[0]), gap)
    col_indice = range(0, len(imgs[0][0]), gap)
    for c in col_indice:
        crop_img = imgs[:, row_indice[1]:row_indice[1]+gap, c: min(len(imgs[0][0]) - gap, c+gap)]
        if crop_img.shape[2] == 0:
            break
        crop_means = np.mean(crop_img, axis=(1, 2))
        crop_stds = np.std(crop_img, axis=(1, 2))
        ret_img_stack = []
        for m, std in zip(crop_means, crop_stds):
            ret_img_stack.append(np.random.normal(loc=m, scale=std, size=(gap, min(len(imgs[0][0]) - gap, c+gap) - c)))
        ret_img_stack = np.array(ret_img_stack)
        imgs[:, row_indice[0]:row_indice[0] + gap, c: min(len(imgs[0][0]) - gap, c+gap)] = np.array(ret_img_stack)
    for r in row_indice:
        crop_img = imgs[:, r:min(len(imgs[0]) - gap, r + gap), len(imgs[0][0]) - 2*gap: len(imgs[0][0]) - gap]
        if crop_img.shape[1] == 0:
            break
        crop_means = np.mean(crop_img, axis=(1, 2))
        crop_stds = np.std(crop_img, axis=(1, 2))
        ret_img_stack = []
        for m, std in zip(crop_means, crop_stds):
            ret_img_stack.append(np.random.normal(loc=m, scale=std, size=(min(len(imgs[0])-gap, r + gap) - r, gap)))
        ret_img_stack = np.array(ret_img_stack)
        imgs[:, r:min(len(imgs[0])-gap, r + gap), len(imgs[0][0]) - gap: len(imgs[0][0])] = np.array(ret_img_stack)
    for c in col_indice[::-1]:
        crop_img = imgs[:, len(imgs[0]) - 2*gap:len(imgs[0]) - gap, c: min(len(imgs[0][0]), c+gap)]
        if crop_img.shape[2] == 0:
            continue
        crop_means = np.mean(crop_img, axis=(1, 2))
        crop_stds = np.std(crop_img, axis=(1, 2))
        ret_img_stack = []
        for m, std in zip(crop_means, crop_stds):
            ret_img_stack.append(np.random.normal(loc=m, scale=std, size=(gap, min(len(imgs[0][0]), c+gap) - c)))
        ret_img_stack = np.array(ret_img_stack)
        imgs[:, len(imgs[0]) - gap:len(imgs[0]), c: min(len(imgs[0][0]), c+gap)] = np.array(ret_img_stack)
    for r in row_indice:
        crop_img = imgs[:, r:min(len(imgs[0]), r + gap), col_indice[1]: col_indice[1] + gap]
        crop_means = np.mean(crop_img, axis=(1, 2))
        crop_stds = np.std(crop_img, axis=(1, 2))
        ret_img_stack = []
        for m, std in zip(crop_means, crop_stds):
            ret_img_stack.append(np.random.normal(loc=m, scale=std, size=(min(len(imgs[0]), r + gap) - r, gap)))
        ret_img_stack = np.array(ret_img_stack)
        imgs[:, r:min(len(imgs[0]), r + gap), col_indice[0]: col_indice[0] + gap] = np.array(ret_img_stack)

    for c in col_indice[1:-1]:
        csize = min(len(imgs[0][0]), c + 2 * gap) - c - gap
        crop_img = (imgs[:, row_indice[0]:row_indice[0]+gap, c-csize: c]
                    + imgs[:, row_indice[0]:row_indice[0]+gap, c+gap: c+gap+csize]) / 2
        crop_means = np.mean(crop_img, axis=(1, 2))
        crop_stds = np.std(crop_img, axis=(1, 2))
        ret_img_stack = []
        for m, std in zip(crop_means, crop_stds):
            ret_img_stack.append(np.random.normal(loc=m, scale=std, size=(gap, min(len(imgs[0][0]), c+gap) - c)))
        ret_img_stack = np.array(ret_img_stack)
        imgs[:, row_indice[0]:row_indice[0] + gap, c: min(len(imgs[0][0]), c+gap)] = np.array(ret_img_stack)
    for r in row_indice[1:-1]:
        rsize = min(len(imgs[0]), r + 2 * gap) - r - gap
        crop_img = (imgs[:, r - rsize: r, len(imgs[0][0]) - 2*gap: len(imgs[0][0]) - gap]
                    + imgs[:, r + gap: r+gap+rsize, len(imgs[0][0]) - 2*gap: len(imgs[0][0]) - gap]) / 2
        crop_means = np.mean(crop_img, axis=(1, 2))
        crop_stds = np.std(crop_img, axis=(1, 2))
        ret_img_stack = []
        for m, std in zip(crop_means, crop_stds):
            ret_img_stack.append(np.random.normal(loc=m, scale=std, size=(min(len(imgs[0]), r + gap) - r, gap)))
        ret_img_stack = np.array(ret_img_stack)
        imgs[:, r:min(len(imgs[0]), r + gap), len(imgs[0][0]) - gap: len(imgs[0][0])] = np.array(ret_img_stack)
    for c in col_indice[1:-1]:
        csize = min(len(imgs[0][0]), c + 2 * gap) - c - gap
        crop_img = (imgs[:, len(imgs[0]) - 2*gap:len(imgs[0]) - gap, c-csize: c]
                    + imgs[:, len(imgs[0]) - 2*gap:len(imgs[0]) - gap, c+gap: c+gap+csize]) / 2
        crop_means = np.mean(crop_img, axis=(1, 2))
        crop_stds = np.std(crop_img, axis=(1, 2))
        ret_img_stack = []
        for m, std in zip(crop_means, crop_stds):
            ret_img_stack.append(np.random.normal(loc=m, scale=std, size=(gap, min(len(imgs[0][0]), c+gap) - c)))
        ret_img_stack = np.array(ret_img_stack)
        imgs[:, len(imgs[0]) - gap:len(imgs[0]), c: min(len(imgs[0][0]), c+gap)] = np.array(ret_img_stack)
    for r in row_indice[1:-1]:
        rsize = min(len(imgs[0]), r + 2 * gap) - r - gap
        crop_img = (imgs[:, r - rsize: r, col_indice[0]: col_indice[0] + gap]
                    + imgs[:, r + gap: r+gap+rsize, col_indice[0]: col_indice[0] + gap]) / 2
        crop_means = np.mean(crop_img, axis=(1, 2))
        crop_stds = np.std(crop_img, axis=(1, 2))
        ret_img_stack = []
        for m, std in zip(crop_means, crop_stds):
            ret_img_stack.append(np.random.normal(loc=m, scale=std, size=(min(len(imgs[0]), r + gap) - r, gap)))
        ret_img_stack = np.array(ret_img_stack)
        imgs[:, r:min(len(imgs[0]), r + gap), col_indice[0]: col_indice[0] + gap] = np.array(ret_img_stack)
    return imgs


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True) 
@cython.profile(False)
cpdef float[:,:,::1] likelihood(crop_imgs, float[:, ::1] gauss_grid, float[::1] bg_squared_sums, bg_means, int window_size1, int window_size2):
    cdef int surface_window, index
    cdef float g_squared_sum, g_mean
    cdef Py_ssize_t i, j
    g_bar = np.zeros([window_size1 * window_size2], dtype=np.float32)
    cdef float[::1] g_bar_view = g_bar 
    surface_window = window_size1 * window_size2
    g_mean = contig_image_mean(gauss_grid)
    g_squared_sum = 0

    index = 0
    for i in range(window_size1):
        for j in range(window_size2):
            g_bar_view[index] = gauss_grid[i][j] - g_mean
            index += 1

    for i in range(window_size1 * window_size2):
        g_squared_sum += g_bar_view[i] * g_bar_view[i]

    i_hat = (crop_imgs - bg_means.reshape(crop_imgs.shape[0], 1, 1))

    i_local_mins = np.min(i_hat, axis=(1, 2))
    #i_hat_mins = np.maximum(np.zeros_like(i_hat_mins, dtype=np.float32), i_hat_mins).reshape(i_hat.shape[0], 1, 1)
    #i_hat = i_hat - i_hat_mins
    for i in range(i_hat.shape[0]):
        i_hat[i,:,:] -= max(0.0, i_local_mins[i])

    i_hat = i_hat @ g_bar / g_squared_sum
    i_hat = np.maximum(np.zeros(i_hat.shape, dtype=np.float32), i_hat)
    L = ((surface_window / 2.) * np.log(1 - (i_hat ** 2 * g_squared_sum).T /
                                        (bg_squared_sums - (surface_window * bg_means)))).T
    return L.reshape(crop_imgs.shape[0], crop_imgs.shape[1], 1).astype(np.float32)


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True) 
@cython.profile(False)
cpdef float[:,:,::1] image_cropping(extended_imgs, int extend, int window_size0, int window_size1, int shift):
    cropped_imgs = []
    cdef int start_row, end_row, start_col, end_col, row_size, col_size, nb_imgs
    cdef int [::1] row_indice, col_indice
    cdef Py_ssize_t n, r, c, index
    nb_imgs = extended_imgs.shape[0]
    row_size = extended_imgs.shape[1]
    col_size = extended_imgs.shape[2]
    start_row = int(extend/2 - (window_size1-1)/2)
    end_row = row_size - window_size1 - start_row + 1
    start_col = int(extend/2 - (window_size0-1)/2)
    end_col = col_size - window_size0 - start_col + 1
    row_indice = np.arange(start_row, end_row, shift, dtype=np.intc)
    col_indice = np.arange(start_col, end_col, shift, dtype=np.intc)

    cropped_imgs = np.empty([nb_imgs, len(row_indice) * len(col_indice), window_size0 * window_size1], dtype=np.float32)
    index = 0
    for r in row_indice:
        for c in col_indice:
            cropped_imgs[:, index] = extended_imgs[:, r:r + window_size1, c:c + window_size0].reshape(-1, window_size0 * window_size1)
            index += 1
    return cropped_imgs


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True) 
@cython.profile(False)
cpdef float[:,:,::1] mapping(float[:,:,::1] c_likelihood, int nb_img, int row_shape, int col_shape, int shift):
    cdef int index
    cdef float[:, :, ::1] c_view = c_likelihood
    cdef float[:, :, ::1] h_view
    h_map = np.zeros([nb_img, row_shape, col_shape], dtype=np.float32)
    h_view = h_map
    if shift == 1:
        return np.array(c_likelihood).reshape(nb_img, row_shape, col_shape)
    else:
        for n in range(nb_img):
            index = 0
            for row in range(row_shape):
                for col in range(col_shape):
                    if row % shift == 0 and col % shift == 0:
                        h_view[n][row][col] = c_view[n][index][0]
                        index += 1
        return h_map



import numpy as np


class TrajectoryObj:
    def __init__(self, index, localizations=None, max_pause=1):
        self.index = index
        self.paused_time = 0
        self.max_pause = max_pause
        self.trajectory_tuples = []
        self.localizations = localizations
        self.times = []
        self.closed = False
        rints = np.random.default_rng(self.index).integers(low=0, high=256, size=3)
        self.color = (rints[0]/255., rints[1]/255., rints[2]/255.)
        self.optimality = 0.
        self.positions = []

    def add_trajectory_tuple(self, next_time, next_position):
        assert self.localizations is not None
        self.trajectory_tuples.append((next_time, next_position))
        x, y, z = self.localizations[next_time][next_position][:3]
        self.positions.append([x, y, z])
        self.times.append(next_time)
        self.paused_time = 0

    def get_trajectory_tuples(self):
        return self.trajectory_tuples

    def add_trajectory_position(self, time, x, y, z):
        self.times.append(time)
        self.positions.append([x, y, z])
        self.paused_time = 0

    def get_positions(self):
        return np.array(self.positions)

    def trajectory_status(self):
        return self.closed

    def close(self):
        self.paused_time = 0
        self.closed = True

    def wait(self):
        if self.paused_time == self.max_pause:
            self.close()
            return self.trajectory_status()
        else:
            self.paused_time += 1
            return self.trajectory_status()

    def get_index(self):
        return self.index

    def get_times(self):
        return np.array(self.times)

    def set_color(self, color):
        self.color = color

    def get_color(self):
        return self.color

    def set_trajectory_tuple(self, trajectory):
        self.trajectory_tuples = trajectory
        self.paused_time = 0

    def get_last_tuple(self):
        return self.trajectory_tuples[-1]

    def get_trajectory_length(self):
        return len(self.get_positions())

    def get_paused_time(self):
        return self.paused_time

    def set_optimality(self, val):
        self.optimality = val

    def get_optimality(self):
        return self.optimality

    def get_expected_pos(self, t):
        if len(self.get_times()) < t+1:
            return np.array(self.positions[-1]), None
        else:
            vector = (np.array(self.positions[-1]) - np.array(self.positions[-1 - t])) / t
            return np.array(self.positions[-1]) + (vector * (self.paused_time + 1)), np.sqrt(vector[0]**2 + vector[1]**2)

    def delete(self, cutoff=2):
        if len(self.positions) < cutoff:
            return True
        else:
            return False

    def get_inst_diffusion_coefs(self, time_interval, t_range=None, ndim=2):
        diff_coefs = []
        if t_range is None:
            t_range = [0, len(self.get_positions())]
        considered_positions = self.get_positions()[t_range[0]: t_range[1]]
        considered_times = self.get_times()[t_range[0]: t_range[1]]

        for i in range(len(considered_positions) - 1):
            j = i + 1
            prev_x, prev_y, prev_z = considered_positions[i]
            prev_t = considered_times[i]
            x, y, z = considered_positions[j]
            t = considered_times[j]
            diff_coef = ((x - prev_x) ** 2 + (y - prev_y) ** 2 + (z - prev_z) ** 2) / (2 * ndim * (t - prev_t))
            diff_coefs.append(diff_coef)
        diff_coefs = np.array(diff_coefs)
        diff_coefs_intervals = []
        for i in range(len(diff_coefs)):
            left_idx = i - time_interval//2
            right_idx = i + time_interval//2
            diff_coefs_intervals.append(np.mean(diff_coefs[max(0, left_idx):min(len(diff_coefs), right_idx+1)]))
        return np.array(diff_coefs_intervals)

    def get_trajectory_angles(self, time_interval, t_range=None):
        """
        available only for 2D data.
        """
        if t_range is None:
            t_range = [0, len(self.get_positions())]
        considered_positions = self.get_positions()[t_range[0]: t_range[1]]
        considered_times = self.get_times()[t_range[0]: t_range[1]]

        angles = []
        for i in range(len(considered_positions) - 2):
            prev_x, prev_y, prev_z = considered_positions[i]
            prev_t = considered_times[i]
            x, y, z = considered_positions[i+1]
            t = considered_times[i+1]
            next_x, next_y, next_z = considered_positions[i+2]
            next_t = considered_times[i+2]
            vec_prev_cur = np.array([x - prev_x, y - prev_y, z - prev_z]) / (t - prev_t)
            vec_cur_next = np.array([next_x - x, next_y - y, next_z - z]) / (next_t - t)

            ang = np.arccos((vec_prev_cur @ vec_cur_next) /
                            (np.sqrt(vec_prev_cur[0] ** 2 + vec_prev_cur[1] ** 2 + vec_prev_cur[2] ** 2)
                             * np.sqrt(vec_cur_next[0] ** 2 + vec_cur_next[1] ** 2 + vec_cur_next[2] ** 2)))
            angles.append(ang)
        angles = np.array(angles)

        angles_intervals = []
        for i in range(len(angles)):
            left_idx = i - time_interval//2
            right_idx = i + time_interval//2
            angles_intervals.append(np.mean(angles[max(0, left_idx):min(len(angles), right_idx+1)]))

        # make length equal to length of xy pos
        # angles_intervals.append(0.0)
        # angles_intervals.append(0.0)
        return np.array(angles_intervals)



import cv2
import sys
import os
import numpy as np
import re
from roifile import ImagejRoi
from FreeTrace.module.data_save import write_trajectory, rewrite_localization
from FreeTrace.module.data_load import read_trajectory, read_localization
from FreeTrace.module.image_module import make_whole_img, make_loc_depth_image


def initialization(gpu, reg_model_nums=[], ptype=-1, verbose=False, batch=False):
    TF = False
    cuda = False
    freetrace_path = ''
    freetrace_path += 'FreeTrace'.join(re.split(r'FreeTrace', __file__)[:-1]) + 'FreeTrace'

    if not os.path.exists(f'{freetrace_path}/models/theta_hat.npz'):
        print(f'\n***** Parmeters[theta_hat.npz] are not found for trajectory inference, please contact author for the pretrained models. *****\n')
        sys.exit('**********************************\n')

    if gpu:
        try:
            import cupy as cp
            if cp.cuda.is_available():
                cuda = True
            else:
                cuda = False
            del cp
        except:
            cuda = False
        try:
            import tensorflow as tf
            gpus = tf.config.list_physical_devices('GPU')
            if len(gpus) > 0:
                TF = True
            else:
                TF = False
            del gpus
        except:
            TF = False

    if TF and ptype==1:
        for reg_model_num in reg_model_nums:
            if not os.path.exists(f'{freetrace_path}/models/reg_model_{reg_model_num}.keras'):
                print(f'\n***** reg_model_{reg_model_num}.keras is not found, contact author for the pretrained models. *****')
                print(f'***********  Contacts  ***********')
                sys.exit('**********************************\n')

    if not batch and verbose:
        track_ = True if ptype==1 else False
        print(f'\n******************************** OPTIONS *****************************************')
        if cuda and TF:
            if track_:
                print(f'*********       Tensorflow: O, Tracking with GPU if fBm_mode is ON.      *********')
            else:
                print(f'*********                Cuda: O, Localization with GPU.                 *********')

        elif cuda and not TF:
            if track_:
                print(f'*********  Tensorflow: X, Tracking without GPU if fBm_mode is ON (SLOW). *********')
            else:
                print(f'*********                Cuda: O, Localization with GPU.                 *********')

        elif not cuda and TF:
            if track_:
                print(f'*********       Tensorflow: O, Tracking with GPU if fBm_mode is ON.      *********')
            else:
                print(f'*********           Cuda: X, Localization without GPU (SLOW).            *********')
                
        else:
            if track_:
                print(f'*********  Tensorflow: X, Tracking without GPU if fBm_mode is ON (SLOW). *********')
            else:
                print(f'*********           Cuda: X, Localization without GPU (SLOW).            *********')
        print(f'**********************************************************************************\n')
        
    if batch and verbose:
        print(f'\n******************************** OPTIONS *****************************************')
        if cuda and TF:
            print(f'**********           Cuda: Ok, Tensorflow: Ok, Fast inferences.          *********')

        elif cuda and not TF:
            print(f'**********      Cuda: Ok, Tensorflow: X, Slow inference on Tracking.     *********')

        elif not cuda and TF:
            print(f'**********    Cuda: X, Tensorflow: Ok, Slow inference on Localisation.   *********')
                
        else:
            print(f'**********            Cuda: X, Tensorflow: X, Slow inferences.           *********') 
        print(f'**********************************************************************************\n')
    return cuda, TF


def crop_trace_roi_and_frame(trace_file:str, roi_file:str|None, start_frame:0, end_frame:9999999, option=0, crop_comparison=False):
    """
    Cropping trajectory result with ROI(region of interest) or frames.
    trace_file: video_traces.csv or equivalent format of trajectory result file.
    roi_file: region_of_interest.roi which containes the roi information in pixel.
    start_frame: start frame to crop the trajectory result with frame.
    end_frame: end frame to crop the trajectory result with frame.
    option: 0 -> considers the trajectories stay only inside the ROI. 1 -> incldues all the trajectories passing trough the ROI.
    crop_comparison: boolean to visualize cropped result.
    """

    assert "traces.csv" in trace_file or "loc.csv" in trace_file, "Wrong trajectory file format, result_traces.csv is needed to crop with ROI or frames"
    assert end_frame > start_frame, "The number of end frame must be greater than start frame."
    assert option == 0 or option == 1, "The option must be 0 or 1. 0: consider the trajectories stay only inside the ROI. 1: incldues all the trajectories passed the ROI."

    if "traces.csv" in trace_file:
        trace_ = True
    else:
        trace_ = False

    if roi_file is None:
        contours = None
    elif type(roi_file) is str and len(roi_file) == 0:
        contours = None
    else:
        contours = ImagejRoi.fromfile(roi_file).coordinates().astype(np.int32)

    if trace_:
        filtered_trajectory_list = []
        trajectory_list = read_trajectory(trace_file)
        xmin = 999999
        xmax = -1
        ymin = 999999
        ymax = -1
        start_frame = max(start_frame, 0)
        end_frame = min(end_frame, 9999999)

        for trajectory in trajectory_list:
            skip = 0
            xyz = trajectory.get_positions()
            times = trajectory.get_times()
            xs = xyz[:, 0]
            ys = xyz[:, 1]
            zs = xyz[:, 2]
            xmin = min(np.min(xs), xmin)
            ymin = min(np.min(ys), ymin)
            xmax = max(np.max(xs), xmax)
            ymax = max(np.max(ys), ymax)

            if option == 0:
                if contours is not None:
                    for x, y in zip(xs, ys):
                        masked = cv2.pointPolygonTest(contours, (x, y), False)
                        if masked == -1:
                            skip = 1
                            break
                    
                    if skip == 1:
                        continue

                if times[0] >= start_frame and times[-1] <= end_frame:
                    filtered_trajectory_list.append(trajectory)
            else:
                if contours is not None:
                    for x, y in zip(xs, ys):
                        masked = cv2.pointPolygonTest(contours, (x, y), False)
                        if masked == 1:
                            if times[0] >= start_frame and times[-1] <= end_frame:
                                filtered_trajectory_list.append(trajectory)
                            break

        print(f'cropping info: ROI[{roi_file}],  Frame:[{start_frame}, {end_frame}]')
        print(f'Number of trajectories before filtering:{len(trajectory_list)}, after filtering:{len(filtered_trajectory_list)}')
        write_trajectory(f'{".".join(trace_file.split("traces.csv")[:-1])}cropped_{start_frame}_{end_frame}_traces.csv', filtered_trajectory_list)
        print(f'{".".join(trace_file.split("traces.csv")[:-1])}cropped_{start_frame}_{end_frame}_traces.csv is successfully generated.')

        if crop_comparison:
            dummy_stack = np.empty((1, int(ymax+1), int(xmax+1), 1))
            make_whole_img(trajectory_list, output_dir=f'{".".join(trace_file.split("traces.csv")[:-1])}before_crop.png', img_stacks=dummy_stack)
            make_whole_img(filtered_trajectory_list, output_dir=f'{".".join(trace_file.split("traces.csv")[:-1])}after_crop.png', img_stacks=dummy_stack)
            print(f'{".".join(trace_file.split("traces.csv")[:-1])}before_crop.png is successfully generated.')
            print(f'{".".join(trace_file.split("traces.csv")[:-1])}after_crop.png is successfully generated.')

    else:
        filtered_loc_list = {}
        filtered_loc_infos = {}
        localisations, loc_infos = read_localization(trace_file)
        nb_locs_before_filtering = 0
        nb_locs_after_filtering = 0

        start_frame = max(start_frame, np.min(list(localisations.keys())))
        end_frame = min(end_frame, np.max(list(localisations.keys())))
        
        for frame in localisations.keys():
            filtered_loc_list[frame] = []
            filtered_loc_infos[frame] = []
            xyz = localisations[frame]
            if contours is not None:
                if len(xyz[0]) > 0:
                    for idx, (x, y, z) in enumerate(xyz):
                        masked = cv2.pointPolygonTest(contours, (x, y), False)
                        if masked != -1:
                            if frame >= start_frame and frame <= end_frame:
                                filtered_loc_list[frame].append([x, y, z])
                                filtered_loc_infos[frame].append(loc_infos[frame][idx])
                                nb_locs_after_filtering += 1
                        nb_locs_before_filtering += 1
            filtered_loc_list[frame] = np.array(filtered_loc_list[frame])
            filtered_loc_infos[frame] = np.array(filtered_loc_infos[frame])

        print(f'cropping info: ROI[{roi_file}],  Frame:[{start_frame}, {end_frame}]')
        print(f'Number of localisations before filtering:{nb_locs_before_filtering}, after filtering:{nb_locs_after_filtering}')
        cropped_file_name = f"{'.'.join(trace_file.split('loc.csv')[:-1])}cropped_{start_frame}_{end_frame}_loc.csv"
        rewrite_localization(cropped_file_name, filtered_loc_list, filtered_loc_infos)
        print(f'{cropped_file_name} is successfully generated.')

        if crop_comparison:
            make_loc_depth_image(output_dir=f'{".".join(trace_file.split("loc.csv")[:-1])}after_crop', coords=[cropped_file_name,])
            print(f'{".".join(trace_file.split("loc.csv")[:-1])}after_crop.png is successfully generated.')        


def calibration_3d(xyz_coords, reg_infos, calib_data):
    """
    3D calibration for astigmatism.
    under development.
    """
    for t in range(len(xyz_coords)):
        for particle_idx in range(len(xyz_coords[t])):
            xvar = reg_infos[t][particle_idx][0]
            yvar = reg_infos[t][particle_idx][1]
            z_coord = (yvar - xvar)
            xyz_coords[t][particle_idx][2] = z_coord
    return xyz_coords
